use std::process::{Command, Stdio};
use std::path::{Path};
use color_print::{cprintln};
use clap::{command, arg, Arg, Parser};

fn trim_newline(s: &mut String) {
    if s.ends_with('\n') {
        s.pop();
        if s.ends_with('\r') {
            s.pop();
        }
    }
}

fn update_apt() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_apt = Command::new("which")
        .arg("apt").stdout(Stdio::piped()).output().unwrap();
    let mut apt_bin = String::from_utf8(find_apt.stdout).unwrap();
    trim_newline(&mut apt_bin);

    let path = Path::new(&apt_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating via: apt</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&apt_bin).arg("update").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&apt_bin).arg("upgrade").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
}

fn update_dnf() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_dnf5 = Command::new("which")
    .arg("dnf5").stdout(Stdio::piped()).output().unwrap();
    let mut dnf5_bin = String::from_utf8(find_dnf5.stdout).unwrap();
    trim_newline(&mut dnf5_bin);

    let find_dnf = Command::new("which")
        .arg("dnf").stdout(Stdio::piped()).output().unwrap();
    let mut dnf_bin = String::from_utf8(find_dnf.stdout).unwrap();
    trim_newline(&mut dnf_bin);

    let path_dnf = Path::new(&dnf_bin);
    let path_dnf5 = Path::new(&dnf5_bin);

    if path_dnf5.exists() {
        cprintln!("<bold><rev>dnf found but skipped since dnf5 is present</rev></bold>");
        cprintln!("<bold><rev>Updating via: dnf5</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&dnf5_bin).arg("--refresh").arg("--nobest").arg("-y").arg("update")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    } else if path_dnf.exists(){
        cprintln!("<bold><rev>Updating via: dnf</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&dnf_bin).arg("--refresh").arg("--skip-broken").arg("--nobest").arg("-y").arg("update")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }

}

fn update_pacman() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_pacman = Command::new("which")
        .arg("pacman").stdout(Stdio::piped()).output().unwrap();
    let mut pacman_bin = String::from_utf8(find_pacman.stdout).unwrap();
    trim_newline(&mut pacman_bin);

    let path = Path::new(&pacman_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating via: pacman</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&pacman_bin).arg("-Syu")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }

}

fn update_urpmi() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_urpmi = Command::new("which")
        .arg("urpmi").stdout(Stdio::piped()).output().unwrap();
    let mut urpmi_bin = String::from_utf8(find_urpmi.stdout).unwrap();
    trim_newline(&mut urpmi_bin);

    let path = Path::new(&urpmi_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating via urpmi</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&urpmi_bin).arg("--auto-update").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
            let _output = cmd.wait();
    }
}

fn update_zypper() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_zypper = Command::new("which")
        .arg("zypper").stdout(Stdio::piped()).output().unwrap();
    let mut zypper_bin = String::from_utf8(find_zypper.stdout).unwrap();
    trim_newline(&mut zypper_bin);

    let path = Path::new(&zypper_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating via: zypper</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&zypper_bin).arg("up").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
}

fn update_flatpak() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_flatpak = Command::new("which")
        .arg("flatpak").stdout(Stdio::piped()).output().unwrap();
    let mut flatpak_bin = String::from_utf8(find_flatpak.stdout).unwrap();
    trim_newline(&mut flatpak_bin);

    let path = Path::new(&flatpak_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating user flatpak installs</rev></bold>");
        let mut cmd =
            Command::new(&flatpak_bin)
            .arg("update").arg("--user").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
    let path = Path::new(&flatpak_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating system flatpak installs</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&flatpak_bin).arg("update").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
}

fn update_snap() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_snap = Command::new("which")
        .arg("snap").stdout(Stdio::piped()).output().unwrap();
    let mut snap_bin = String::from_utf8(find_snap.stdout).unwrap();
    trim_newline(&mut snap_bin);

    let path = Path::new(&snap_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating user snap installation</rev></bold>");
        let mut cmd =
            Command::new(&snap_bin)
            .arg("refresh").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
    let path = Path::new(&snap_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating system snap installation</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&snap_bin).arg("refresh").arg("-y")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
}

fn update_python() {
    let find_pip_review = Command::new("which")
        .arg("pip-review").stdout(Stdio::piped()).output().unwrap();
    let mut pip_review_bin = String::from_utf8(find_pip_review.stdout).unwrap();
    trim_newline(&mut pip_review_bin);

    let path = Path::new(&pip_review_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating Python user installation via pip-review</rev></bold>");
        let mut cmd =
            Command::new(&pip_review_bin)
            .arg("--auto").arg("--local").arg("--user")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
}

fn update_distrobox() {
    let find_distrobox_upgrade = Command::new("which")
        .arg("distrobox-upgrade").stdout(Stdio::piped()).output().unwrap();
    let mut distrobox_upgrade_bin = String::from_utf8(find_distrobox_upgrade.stdout).unwrap();
    trim_newline(&mut distrobox_upgrade_bin);

    let path = Path::new(&distrobox_upgrade_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating Distrobox containers</rev></bold>");
        let mut cmd =
            Command::new(&distrobox_upgrade_bin)
            .arg("--all")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
}

fn update_podman() {
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    let find_podman = Command::new("which")
        .arg("podman").stdout(Stdio::piped()).output().unwrap();
    let mut podman_bin = String::from_utf8(find_podman.stdout).unwrap();
    trim_newline(&mut podman_bin);

    let path = Path::new(&podman_bin);
    if path.exists(){
        cprintln!("<bold><rev>Updating Podman system containers</rev></bold>");
        let mut cmd =
            Command::new(&sudo_bin)
            .arg(&podman_bin).arg("auto-update")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }

    if path.exists(){
         cprintln!("<bold><rev>Updating Podman user containers</rev></bold>");
         let mut cmd =
             Command::new(&podman_bin)
             .arg("auto-update")
             .stdout(Stdio::inherit()).stderr(Stdio::inherit())
             .spawn().unwrap();
         let _output = cmd.wait();
     }

     if path.exists(){
         cprintln!("<bold><rev>Pruning Podman system containers</rev></bold>");
         let mut cmd =
             Command::new(&sudo_bin)
             .arg(&podman_bin).arg("system").arg("prune").arg("-af")
             .stdout(Stdio::inherit()).stderr(Stdio::inherit())
             .spawn().unwrap();
         let _output = cmd.wait();
     }

    if path.exists(){
        cprintln!("<bold><rev>Pruning Podman user containers</rev></bold>");
        let mut cmd =
            Command::new(&podman_bin)
            .arg("system").arg("prune").arg("-af")
            .stdout(Stdio::inherit()).stderr(Stdio::inherit())
            .spawn().unwrap();
        let _output = cmd.wait();
    }
}

fn update_all() {
    // Tell user that we are going to start the update procedure
    cprintln!("<bold><rev>osupdater : This tool will locate and run any package/container managers on your system.</rev></bold>\n");

    // Check package managers
    cprintln!("<bold><rev>Looking for native package managers.</rev></bold>\n");
    update_apt();
    update_dnf();
    update_pacman();
    update_urpmi();
    update_zypper();

    // Check container formats
    cprintln!("<bold><rev>Checking application containers</rev></bold>");
    update_flatpak();
    update_snap();
    update_python();
    update_distrobox();
    update_podman();
}


fn main() {

    // Clear termianl, and move cursor to 1,1 inside the terminal
    // print!("{esc}[2J{esc}[1;1H", esc = 27 as char);

    // Find sudo
    let find_sudo = Command::new("which")
        .arg("sudo").stdout(Stdio::piped()).output().unwrap();
    let mut sudo_bin = String::from_utf8(find_sudo.stdout).unwrap();
    trim_newline(&mut sudo_bin);

    // update_all();


    let matches = command!() // requires `cargo` feature
            .arg(
                arg!(<UPDATE_TOOL>)
                    .help("What updater to run")
                    .value_parser(["all", "apt", "dnf", "pacman", "urpmi", "zypper", "flatpak", "snap", "python", "distrobox", "podman"])
                    .required(false)
                    .default_value("all"),
            )
            .get_matches();

        // Note, it's safe to call unwrap() because the arg is required
        match matches
            .get_one::<String>("UPDATE_TOOL")
            .expect("'UPDATE_TOOL' is required, else all will be run.")
            .as_str()
        {
            "all" => {
                update_all();
            }

            "apt" => {
                update_apt();
            }
            "dnf" => {
                update_dnf();
            }
            "pacman" => {
                update_pacman();
            }
            "urpmi" => {
                update_urpmi();
            }
            "zypper" => {
                update_zypper();
            }
            "flatpak" => {
                update_flatpak();
            }
            "snap" => {
                update_snap();
            }
            "python" => {
                update_python();
            }
            "distrobox" => {
                update_distrobox();
            }
            "podman" => {
                update_podman();
            }
            _ => unreachable!(),
        }


}
